import asyncio
import logging.config
import os
import sys
from collections.abc import Callable, Coroutine
from functools import wraps
from json import JSONDecodeError
from pathlib import Path
from typing import Any, TypeVar

import click
import uvicorn
from botocore.exceptions import ClientError, NoCredentialsError
from pydantic import ValidationError

from sagemaker_shim.app import app
from sagemaker_shim.logging import LOGGING_CONFIG
from sagemaker_shim.models import InferenceTaskList, get_s3_file_content

T = TypeVar("T")


def async_to_sync(
    func: Callable[..., Coroutine[Any, Any, T]]
) -> Callable[..., T]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


def _get_writable_directories() -> list[str]:
    directories_str = os.environ.get(
        "GRAND_CHALLENGE_COMPONENT_WRITABLE_DIRECTORIES", ""
    )
    return [d for d in directories_str.split(":") if d != ""]


def _ensure_directories_are_writable() -> None:
    for directory in _get_writable_directories():
        path = Path(directory)
        path.mkdir(exist_ok=True, parents=True)
        path.chmod(mode=0o777)


@click.group()
def cli() -> None:
    logging.config.dictConfig(LOGGING_CONFIG)
    _ensure_directories_are_writable()


@cli.command(short_help="Start the model server")
def serve() -> None:
    uvicorn.run(app=app, host="0.0.0.0", port=8080, log_config=None, workers=1)


@cli.command(short_help="Invoke the model")
@click.option(
    "-t",
    "--tasks",
    default=None,
    help="A JSON string of task definitions",
    type=str,
)
@click.option(
    "-f",
    "--file",
    default=None,
    help="S3 URI of a JSON file containing a list of task definitions",
    type=str,
)
@async_to_sync
async def invoke(tasks: str, file: str) -> None:
    tasks_json: str | bytes

    if tasks and file:
        raise click.UsageError("Only one of tasks or file should be set")
    elif tasks:
        tasks_json = tasks
    elif file:
        try:
            tasks_json = get_s3_file_content(s3_uri=file)
        except (ValueError, ClientError, NoCredentialsError) as error:
            raise click.BadParameter(
                f"The value provided for file is invalid:\n\n{error}"
            ) from error
    else:
        raise click.UsageError("One of tasks or file should be set")

    try:
        parsed_tasks = InferenceTaskList.model_validate_json(tasks_json)
    except (ValidationError, JSONDecodeError) as error:
        raise click.BadParameter(
            f"The tasks definition is invalid:\n\n{error}"
        ) from error

    if not parsed_tasks.root:
        raise click.UsageError("Empty task list provided")

    for task in parsed_tasks.root:
        # Only run one task at a time
        await task.invoke()


if __name__ == "__main__":
    # https://pyinstaller.org/en/stable/runtime-information.html#run-time-information
    we_are_bundled = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")

    if we_are_bundled:
        # https://pyinstaller.org/en/stable/runtime-information.html#using-sys-executable-and-sys-argv-0
        cli(sys.argv[1:])
    else:
        cli()



import json
import logging
import os
from typing import Any

logger = logging.getLogger(__name__)

STDOUT_LEVEL = logging.INFO


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        """
        Create a structured log message

        CloudWatch does not separate log streams, so we use source
        to differentiate between stdout and stderr.
        Internal allows us to differentiate between logs from
        this application, and logs from the invocation subprocess.
        """
        message = super().format(record=record)

        # We need to explicitly add the source as an annotation
        # as CloudWatch unifies the logs
        if record.levelno <= STDOUT_LEVEL:
            source = "stdout"
        else:
            source = "stderr"

        internal = getattr(record, "internal", True)
        task = getattr(record, "task", None)
        task_pk = getattr(task, "pk", None)

        return "\n".join(
            json.dumps(
                {
                    "log": m,
                    "level": record.levelname,
                    "source": source,
                    "internal": internal,
                    "task": task_pk,
                }
            )
            for m in message.splitlines()
        )


class StdStreamFilter(logging.Filter):
    """Split stdout and stderr streams"""

    def __init__(self, *args: Any, stdout: bool, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__stdout = stdout

    def filter(self, record: logging.LogRecord) -> bool:
        """Should this log message be displayed?"""
        if self.__stdout:
            # stdout, STDOUT_LEVEL and lower
            return record.levelno <= STDOUT_LEVEL
        else:
            # stderr, greater than STDOUT_LEVEL
            return record.levelno > STDOUT_LEVEL


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "stdout": {
            "()": StdStreamFilter,
            "stdout": True,
        },
        "stderr": {
            "()": StdStreamFilter,
            "stdout": False,
        },
    },
    "formatters": {
        "json": {
            "()": JSONFormatter,
        },
    },
    "handlers": {
        "console_stdout": {
            "class": "logging.StreamHandler",
            "formatter": "json",
            "stream": "ext://sys.stdout",
            "filters": ["stdout"],
        },
        "console_stderr": {
            "class": "logging.StreamHandler",
            "formatter": "json",
            "stream": "ext://sys.stderr",
            "filters": ["stderr"],
        },
    },
    "root": {
        "level": os.environ.get("LOG_LEVEL", "INFO").upper(),
        "handlers": ["console_stdout", "console_stderr"],
    },
}



import logging
import re
import zipfile
from os.path import commonpath
from pathlib import Path

from sagemaker_shim.exceptions import ZipExtractionError
from sagemaker_shim.vendor.werkzeug.security import safe_join

logger = logging.getLogger(__name__)


def _filter_members(members: list[zipfile.ZipInfo]) -> list[dict[str, str]]:
    """Filter common paths and uninteresting files from a zip archive"""
    valid_files = [
        m
        for m in members
        if not m.is_dir()
        and re.search(r"(__MACOSX|\.DS_Store|desktop.ini)", m.filename) is None
    ]

    if len(valid_files) == 0:
        # No files present
        return []
    elif len(valid_files) == 1:
        common_directory = str(Path(valid_files[0].filename).parent)
        common_directory = "" if common_directory == "." else common_directory
    else:
        common_directory = commonpath([m.filename for m in valid_files])

    if common_directory:
        sliced_directory = slice(len(common_directory) + 1, None, None)
    else:
        sliced_directory = slice(None, None, None)

    return [
        {"src": f.filename, "dest": f.filename[sliced_directory]}
        for f in valid_files
    ]


def safe_extract(*, src: Path, dest: Path) -> None:
    """
    Safely extracts a zip file into a directory

    Any common prefixes and system files are removed.
    """

    if not dest.exists():
        raise RuntimeError("The destination must exist")

    with src.open("rb") as f:
        with zipfile.ZipFile(f) as zf:
            members = _filter_members(zf.infolist())

            for member in members:
                file_dest = safe_join(str(dest), member["dest"])

                if file_dest is None:
                    raise ZipExtractionError("Zip file contains invalid paths")

                # We know that the dest is within the prefix as
                # safe_join is used, and the destination is already
                # created, so ok to create the parents here
                file_dest.parent.mkdir(exist_ok=True, parents=True)

                logger.info(
                    f"Extracting {member['src']=} from {src} to {file_dest}"
                )

                with zf.open(member["src"], "r") as fs, open(
                    file_dest, "wb"
                ) as fd:
                    while chunk := fs.read(8192):
                        fd.write(chunk)
